import os
from faunanet import repl

if __name__ == "__main__":
    cmd = repl.FaunanetCmd()

    run_cfg = os.environ.get("RUN_CONFIG", "")

    if run_cfg != "":
        cmd.do_start("--cfg=/root/faunanet_config/" + run_cfg)
    else:
        cmd.do_start("")

    cmd.wait_for_watcher_event(
        lambda _: cmd.watcher is not None and cmd.watcher.is_running,
    )

    cmd.do_status("")

    cmd.cmdloop()



import os
import numpy as np
from pathlib import Path
import datetime
import warnings

from birdnetlib.main import RecordingBase
from faunanet.preprocessor_base import PreprocessorBase
from faunanet.model_base import ModelBase
import faunanet.utils as utils
from faunanet.species_predictor import SpeciesPredictorBase


class Recording(RecordingBase):
    """
    Recording A Recording represents an audio recoding of arbitrary length and date associated with a preprocessor and analyzer object.

    Recordings read and preprocess audio data using the supplied preprocessor object and can use the 'analyzer' object to analyze and classify the contained animal sounds

    Args:
        RecordingBase (birdnetlib.RecordingBase): See documentation of the birdnetlib.RecordingBase class
    """

    def _check_model_system_viability(self, names_of_objects: list):

        # README: 'week' argument is not used because it doubles with 'date', and daily timestamps are more useful for data collection in general
        if all(name == names_of_objects[0] for name in names_of_objects) is False:

            raise ValueError(
                "Found different 'name' attributes for model, preprocessor and species predictor. Make sure the supplied model, preprocessor and species predictor are compatible to each other (species_predictor may be 'None' if not used)."
            )

    def __init__(
        self,
        preprocessor: PreprocessorBase,
        model: ModelBase,
        path: str,
        date: datetime.datetime = None,
        lat: float = None,
        lon: float = None,
        species_presence_threshold: float = 0.03,
        min_conf: float = 0.25,
        species_predictor: SpeciesPredictorBase = None,
        file_check_poll_interval: int = 1,
    ):
        # README: The arguments lon, lat, species_presence_threshold and week_48, date should be moved out of the __init__ at some point, at some point perhaps?
        """
        Create a new Recording.

        Args:
            analyzer (Analyzer): Analyzer object to use. Contains model to use for analysis as well as result post processing.
            preprocessor (PreprocessorBase): Preprocessor object to use. Must adhere to the interface defined in faunanet.preprocessor_base.
            path (str): Path to the audio file to be analyzed
            date (datetime, optional): Date of recording. Alternative to 'week'. Defaults to None. Only applied if `model` is the birdnet default model.
            sensitivity (float, optional): Detection sensitivity. Defaults to 1.0.
            lat (float, optional): Latitude. If latitude and longitude are given, the species list is predicted first.  Defaults to None. Only applied if `model` is the birdnet default model.
            lon (float, optional): Longitude. If latitude and longitude are given, the species list is predicted first.  Defaults to None. Only applied if `model` is the birdnet default model.
            species_presence_threshold (float, optional): The threshold for the species predictor. A species predicted to be present above that threshold is added to the list of allowed species, below it is excluded. Restricted to [0,1]
            min_conf (float, optional): Minimal confidence to use to consider a detection valid. Defaults to 0.1.
            species_predictor: An instance of a class derived from `SpeciesPredictorBase`. Only applicable if `model` is the birdnet default model.
            file_check_poll_interval: Inteval in seconds at which the recording should check whether the file to be analyzed is still being written to. Defaults to 1
        """
        self.processor = preprocessor
        self.analyzer = model
        self.path = path
        p = Path(self.path)
        self.filestem = p.stem
        self.species_predictor = None
        self.allowed_species = []
        self.file_check_poll_interval = file_check_poll_interval

        # make sure that all the system components are compatible. Based on name tags. Still a bit susceptible. Fix?

        self._check_model_system_viability(
            [
                model.name,
                preprocessor.name,
                model.name if species_predictor is None else species_predictor.name,
            ]
        )

        species_predictor_args = [lat, lon, species_predictor, date]

        # check that necessary info is present for species predictor: nothing necessary is None and a date variable is given
        if all(var is not None for var in species_predictor_args):

            self.species_predictor = species_predictor

            self.allowed_species = self.species_predictor.predict(
                latitude=lat,
                longitude=lon,
                date=date,
                threshold=species_presence_threshold,
            )

        # when any of the species predictor stuff is None and something else is not,
        # we have an incompatible combination.
        if any(var is None for var in species_predictor_args) and any(
            var is not None for var in species_predictor_args
        ):
            raise ValueError(
                "A full set of (lat, lon, date/week_48) must be given to use the species presence predictor and the passed species predictor must not be 'None'."
            )

        # README: can we get rid of the birdnetlib-Recording thing completely at some point?

        # we ignore the underlying species predictor stuff that the birdnetlib base class does
        # by not passing it the respective arguments.
        super().__init__(
            model,
            sensitivity=1.0,
            min_conf=min_conf,
            overlap=0.0,
            return_all_detections=False,  # not used in faunanet. Set min_conf to 0 for the same effect
        )

    @property
    def filename(self):
        return os.path.basename(self.path)

    @property
    def chunks(self):
        return self.processor.chunks

    def set_analyzer(
        self,
        model: ModelBase,
        preprocessor: PreprocessorBase,
        species_predictor: SpeciesPredictorBase = None,
    ):
        """
        set_analyzer Replace the internally used model and preprocessor.

        Args:
            model (ModelBase): The classifier model to use
            preprocessor (PreprocessorBase): The preprocessor to use
        """

        self._check_model_system_viability(
            [
                model.name,
                preprocessor.name,
                model.name if species_predictor is None else species_predictor.name,
            ]
        )

        self.analyzer = model

        self.processor = preprocessor

        if species_predictor is not None:
            self.species_predictor = species_predictor

        self.analyzed = False

    def restrict_species_list(
        self,
        lat: float,
        lon: float,
        date: datetime.datetime = None,
        species_presence_threshold: float = 0.03,
    ):
        # this abstracts away the underlying species predictor model and is useful when reusing the recording or
        # into whatever class it may morph in the future.
        """
        restrict_species_list Restrict the list of possibly present species based on coordinates and time.

        Args:
            lat (float): Latitude coordinate to predict species presence for.
            lon (float): Longitude coordinate to predict species presence for.
            date (datetime, optional): Date to predict species presence for. Only relevant when 'week' is not -1. Defaults to None.
            week (int, optional): Week (4 weeks per month) to predict species presence for. Defaults to -1. Only relevant when 'date' is None.
            species_presence_threshold (float, optional): Threshold above which a species is considered to be persent. Restricted to [0, 1]. Defaults to 0.03.
        """
        self.allowed_species = self.species_predictor.predict(
            latitude=lat,
            longitude=lon,
            date=date,
            threshold=species_presence_threshold,
        )

    def process_audio_data(self, data: np.ndarray) -> list:
        """Process raw audio data via processor.process_audio_data.

        Args:
            data (np.ndarray): Raw audio data returned from preprocessor.read_audio_data

        Returns:
            list: data processed to be analyzed
        """
        return self.processor.process_audio_data(data)

    def read_audio_data(self):
        """Read audio data from file and pass it on for preprocessing.

        Returns:
            list[np.array]: preprocessed audio data
        """

        # README: wait until the file to be opened does not change in size anymore.
        # only working way in which we do not have to make assumptions about the
        # system that actually writes the file?
        utils.wait_for_file_completion(
            str(self.path), polling_interval=self.file_check_poll_interval
        )

        rawdata = self.processor.read_audio_data(self.path)

        return self.process_audio_data(rawdata)

    @classmethod
    def from_cfg(cls, faunanet_path: str, cfg: dict):
        """
        from_cfg Create a new Recording from a dictionary containing all keyword arguments. Usually, this is obtained by reading in a YAML config.

        Args:
            faunanet_path (str): Path to the faunanet installation
            cfg (dict): keyword arguments for the Recording and its 'preprocessor' and 'model' attributes.

        Returns:
            Recording: New instance build with supplied kwargs.
        """
        # README: faunanet path needed still -> can we get rid of it in some way?
        # config.py/.yml written upon install or something....
        # future PR when installing/packaging is done

        # load appropriate modules: preprocessor, model

        module_path = Path(faunanet_path) / Path("models") / cfg["Model"]["model_path"]

        preproc_m = utils.load_module("pp", str(module_path / "preprocessor.py"))

        model_m = utils.load_module("mm", str(module_path / "model.py"))

        # make preprocessor instance
        preprocessor = preproc_m.Preprocessor.from_cfg(cfg["Data"]["Preprocessor"])

        # make model instance
        model = model_m.Model.from_cfg(cfg["Analysis"]["Model"])

        # make recording instance and return

        defaults = {
            "path": None,
            "week_48": -1,
            "date": None,
            "lat": None,
            "lon": None,
            "return_all_detections": False,
            "min_conf": 0.25,
        }

        del cfg["Analysis"]["Model"]

        return cls(preprocessor, model, **(defaults | cfg["Analysis"]))

    @property
    def detections(self):
        """
        detections Produce a list of tuples containting (start_time, end_time, label, detection_confidence) from the raw detections produced by the 'model' attribute of this class.
        Also filters them by minimum confidence level.

        Returns:
            List(Tuples): A list containing a tuple (start, end, lable, confidence) for each detection with confidence > self.minimum_confidence.
        """
        # Readme: overrides the detections method of the base class to make things a bit simpler and remove stuff
        # that is currently not supported here or which is too specific.

        if not self.analyzed:
            warnings.warn(
                "'analyze' method has not been called. Call .analyze() before accessing detections.",
                RuntimeWarning,
            )

        qualified_detections = []

        allow_list = set(
            self.allowed_species
        )  # convert to set for lookup speed. Useful? (maybe for bigger lists...)

        for (start, end), labeled_predictions in self.analyzer.results.items():

            for label, confidence in labeled_predictions:
                if confidence > self.minimum_confidence:
                    if label in allow_list or len(allow_list) == 0:

                        qualified_detections.append(
                            {
                                "start": start,
                                "end": end,
                                "label": label,
                                "confidence": confidence,
                            }
                        )

        return qualified_detections



from platformdirs import user_cache_dir
from pathlib import Path
import datetime
from birdnetlib.species import SpeciesList
from birdnetlib.utils import return_week_48_from_datetime
from shutil import rmtree
import os
from . import utils


# README: Would be more logically consistent to have this inherit from faunanet's `ModelBase` class,
# but that would require a lot of copy-cat code from birdnetlib.


class SpeciesPredictorBase(SpeciesList):

    def __init__(
        self,
        model_path: str,
        use_cache: bool = True,
        num_threads: int = 1,
    ):
        """
        __init__ Create a new SpeciesPredictorBase instance.

        Args:
            model_path (str): _description_
            use_cache (bool, optional): _description_. Defaults to True.
            num_threads (int, optional): _description_. Defaults to 1.
        """

        if Path(model_path).exists() is False:
            raise FileNotFoundError(
                "The model folder for the species presenece predictor could not be found"
            )

        labels_path = Path(model_path) / "labels.txt"

        if labels_path.exists() is False:
            raise FileNotFoundError(
                "The 'labels' file for the species presence predictor could not be found"
            )

        model_path = Path(model_path) / "species_presence_model.tflite"

        if model_path.exists() is False:
            raise FileNotFoundError(
                "The 'model' file for the species presence predictor could not be found"
            )

        self.model_path = str(model_path)

        self.labels_path = str(labels_path)

        self.use_cache = use_cache

        self.num_threads = num_threads

        self.results = []

        self.read_from_file = False  # variable used for debugging mostly

        self.name = "birdnet_default"

        if os.getenv("FAUNANET_TEST_MODE") == "True":
            self.cache_dir = (
                user_cache_dir() / Path("faunanet_tests") / Path("species_lists")
            )
        else:
            self.cache_dir = user_cache_dir() / Path("faunanet") / Path("species_lists")

        if self.cache_dir.exists() is False:
            self.cache_dir.mkdir(parents=True, exist_ok=True)

        super().__init__()  # super class handles model loading

    def load_species_list_model(self):
        """
        load_species_list_model Load the default species presence model used by Birdnet-Analyzer.
        """
        self.meta_interpreter = utils.load_model_from_file_tflite(
            self.model_path, self.num_threads
        )
        self.meta_interpreter.allocate_tensors()
        # Get input and output tensors.
        self.meta_input_details = self.meta_interpreter.get_input_details()
        self.meta_output_details = self.meta_interpreter.get_output_details()

        # Get input tensor index
        self.meta_input_layer_index = self.meta_input_details[0]["index"]
        self.meta_output_layer_index = self.meta_output_details[0]["index"]

    def load_labels(self):
        """
        load_species_list_model Load the full species labels list used by the prediction model that should be restricted.
        """
        # README: this assumes there is only one label per line and nothing else -> fix by using pandas?
        self.labels = self._read_labels_file(self.labels_path)

    def predict_species_list(
        self,
        latitude: float,
        longitude: float,
        date: datetime.datetime = None,
        threshold: float = 0.03,
    ):
        """
        predict_species_list Apply the model for species presence prediction at (longitude, latitude) coordinates at date 'date' (will be converted to a week) or within a given week, using a confidence threshold to mark species as present. Override this in a derived  class if you want to customize how species presence is restricted.

        Args:
            latitude (float): Latitude value for location coordinate
            longitude (float): Longitude value for location coordinate
            date (datetime.datetime, optional): Date to restrict the species presence to. If both 'week' and 'date' are present, the week will be computed from the date. Defaults to None.
            threshold (float, optional): Detection threshold to mark a species as being present. Defaults to 0.03.
        """
        self.results = []

        # this relies on the birdnetlib format and returns [{
        #     "scientific_name": sc,
        #     "common_name": cn,
        #     "threshold": t,
        # }, {"scientific_name: " sc, ...}, { ... }]

        self.results = self.return_list(
            lon=longitude,
            lat=latitude,
            date=date,
            threshold=threshold,
        )

    # README: coordinates are not members becaues this is intended to be used from within another class that holds those
    def predict(
        self,
        latitude: float,
        longitude: float,
        date: datetime.datetime = None,
        threshold: float = 0.03,
    ) -> list:
        """
        predict Predict the presence of species in the original label set using coordinates (latitude, longitude) and a timestep (week within a year), using a supplied threshold to set a confidence limit to mark a species as present. if `use_cache` is true in the caller, attempts to use a cached species list instead of computing a new one.
        Otherwise, the arguments are passed internally to the 'predict_species_list' function which then creates a new list. Override in a derived class to customize behavior.
        Args:
            latitude (float): Latitude value for location coordinate
            longitude (float): Longitude value for location coordinate
            date (datetime.datetime, optional): Date to restrict the species presence to. If both 'week' and 'date' are present, the week will be computed from the date. Defaults to None.
            week (int, optional): Week to restrict the species presence. If both 'week' and 'date' are present, the week will be computed from the date. Defaults to -1.
            threshold (float, optional): Detection threshold to mark a species as being present. Defaults to 0.03.

        Raises:
            ValueError: When week and date are both None. One of both needs to be present.

        Returns:
            list: List of label names according to the formatting applied in the 'detections' function.
        """

        self.read_from_file = False

        if date is not None:

            week = return_week_48_from_datetime(
                date
            )  # not optimal, this is called in ´return_results' again.

        # when we want to use cached results, we first check if we computed a list for this location and date before
        cache_folder = Path(str(latitude) + "_" + str(longitude) + "_" + str(week))

        if (
            self.use_cache
            and (self.cache_dir / cache_folder / "species_list.txt").exists()
        ):

            self.read_from_file = True

            return self._read_labels_file(
                Path(self.cache_dir) / cache_folder / "species_list.txt"
            )

        # provides raw species presence data in self.result
        self.predict_species_list(
            latitude=latitude,
            longitude=longitude,
            date=date,
            threshold=threshold,
        )

        detections = self.detections

        if self.use_cache:  # store when requested. Round given coords to 2 decimals

            self._write_to_file(
                Path(self.cache_dir) / cache_folder,
                detections,
            )

        return detections

    @property
    def detections(self) -> list:
        """
        detections Turn raw species presences prediction results into the labels format used by the model. Override this method in a derived class to get custom formats for different models

        Returns:
            list: Predicted species in the format needed by the model
        """
        # transform the species list into the birdnet format.

        detections = []
        for dictionary in self.results:
            sn = dictionary["scientific_name"]
            cn = dictionary["common_name"]

            detections.append(f"{sn}_{cn}")

        return detections

    def export(self, outpath: str):
        """
        export Export a generated species list to file residing at 'outputpath'

        Args:
            outpath (str): absolute path to the output file.

        Raises:
            RuntimeError: When there is no stored specieslist to write to file.
        """
        if len(self.results) == 0:
            raise RuntimeError(
                "No results present, first generate a species list by calling 'predict'"
            )

        self._write_to_file(Path(outpath), self.detections)

    def _read_labels_file(self, filepath: Path) -> list:
        """
        _read_labels_file Reads labels.txt file. Only single-column files are supported.

        """
        with open(filepath, "r") as lfile:
            read_data = [line.replace("\n", "") for line in lfile.readlines()]

        return read_data

    def _write_to_file(self, containing_folder: Path, formatted_results: list):
        """Write produced species list to 'containing_folder' within the cache_dir path held by the caller. Creates directories as necessary"""

        containing_folder.mkdir(parents=True, exist_ok=True)

        with open(containing_folder / "species_list.txt", "w") as output:
            for row in formatted_results:
                output.write(row + "\n")

    def clear_cache(self):
        """
        clear_cache Remove the cache folder completely

        """
        if self.cache_dir.exists():
            rmtree(str(self.cache_dir))



from .model_base import ModelBase
from .recording import Recording
from .preprocessor_base import PreprocessorBase
from .species_predictor import SpeciesList
from .species_predictor import SpeciesPredictorBase
from .watcher import Watcher

__all__ = ["ModelBase", "Recording", "PreprocessorBase", "SpeciesList", "SpeciesPredictorBase", "Watcher", "__version__"]
__version__ = "0.0.10"



from abc import ABC, abstractmethod
import numpy as np
import librosa
import audioread


class AudioFormatError(Exception):
    pass


class PreprocessorBase(ABC):
    """
    PreprocessorBase A common interface for all custom preprocessor classes

    """

    def __init__(
        self,
        name: str,
        sample_rate: int = 32000,
        sample_secs: float = 5.0,
        overlap: float = 0.0,
        resample_type: str = "kaiser_fast",
    ):
        self.name = name
        self.sample_rate = sample_rate
        self.overlap = overlap
        self.sample_secs = sample_secs
        self.resample_type = resample_type
        self.duration = 0
        self.actual_sampling_rate = 0
        self.chunks = []

    def read_audio_data(self, path: str) -> np.array:
        """
        read_audio_data Read in audio data, resample and return the resampled raw data, adding members for actual sampling rate and duration of audio file.
        Args:
            path (str): Path to the audio file to be analyzed

        Raises:
            AudioFormatError: When the format of the audio file is unknown

        Returns:
            np.ndarray: resampled raw data
        """

        try:
            data, rate = librosa.load(
                path, sr=self.sample_rate, mono=True, res_type=self.resample_type
            )

            self.duration = librosa.get_duration(y=data, sr=self.sample_rate)
            self.actual_sampling_rate = rate

        except audioread.exceptions.NoBackendError as e:
            raise AudioFormatError("Audio format could not be opened.") from e
        except FileNotFoundError as e:
            raise e
        except Exception as e:
            raise AudioFormatError(
                "Generic audio read error occurred from librosa."
            ) from e

        if self.actual_sampling_rate != self.sample_rate:
            raise RuntimeError(
                "Error, sampling rate from resampling and desired sampling rate don't match"
            )

        print(
            f"read audio data of duration {self.duration} from {path} with sampling rate {self.actual_sampling_rate}"
        )

        return data

    @abstractmethod
    def process_audio_data(self, rawdata: np.ndarray) -> list:
        """
        process_audio_data Receives raw read data and processes it into something that can be used by the analyzer model.


        Args:
            rawdata (np.ndarray): Raw audiodata received from the read_audio_data function.

        Returns:
            list: final preprocessed audiodata ready to be passed to the analyzer
        """

        pass



from abc import ABC, abstractmethod
import numpy as np
from pathlib import Path
import operator
from . import utils


class ModelBase(ABC):
    """
    ModelBase Base model class for faunanet. Every model that faunanet can use must derive from this class.

    Attributes:
        model_path (str): path to the model file used, or the url used to load it
        labels_path (str): path to the labels file read for having labels
        species_list_path (str|None): Path to a restricted list of species labels if used, else None
        sensitivity (float, defaults to 1.0): Parameter of the sigmoid activation function used to produce classification probabilities.
    """

    def __init__(
        self,
        name: str,
        model_path: str,
        labels_path: str,
        num_threads: int = 1,
        **kwargs,
    ):
        self.num_threads = num_threads

        if Path(model_path).exists() is False:
            raise FileNotFoundError(f"No model file at {model_path}")

        if Path(labels_path).exists() is False:
            raise FileNotFoundError(f"No labels file at {labels_path}")

        self.name = name

        self.model_path = model_path

        self.labels_path = labels_path

        self.results = None

        self.load_model()

        self.load_labels()

    def _sigmoid(self, logits: np.array, sigmoid_sensitivity: float = -1):
        """
        _sigmoid Apply a simple sigmoid to output logits to map them to probabilities

        Args:
            logits (np.array): Raw output from a the inference function of the loaded model.
            sigmoid_sensitivity (float, optional): Sigmoid parameter. Defaults to -1.

        Returns:
            np.array: Model output mapped to [0,1] to get interpretable probability
        """
        return 1 / (1.0 + np.exp(sigmoid_sensitivity * np.clip(logits, -15, 15)))

    def load_labels(self):
        """
        load_labels Read the labels file as a pandas dataframe.
        """
        labels = []
        with open(self.labels_path, "r") as lfile:
            for line in lfile.readlines():
                labels.append(line.replace("\n", ""))
        self.labels = labels
        print("Labels loaded.")

    def load_model(self):
        """
        load_model Load the model to be used for classifying.

        Raises:
            ValueError: When the path to a local model file does not work
            ValueError: When a given url does not refer to a huggingface hub model
        """
        model_loaders = {
            "model.tflite": utils.load_model_from_file_tflite,
            "saved_model.pb": utils.load_model_from_file_pb,
        }

        # get the model file
        model_filename = str(Path(self.model_path).name)

        # load model locally

        if model_filename not in model_loaders:
            raise ValueError(f"Model filename unknown: {model_filename}")

        self.model = model_loaders[model_filename](self.model_path, self.num_threads)

    @abstractmethod
    def predict(self, data: np.array) -> list:
        pass

    def analyze_recording(self, recording):
        """
        analyze_recording _summary_

        Args:
            recording (_type_): _description_
        """
        print("name of model: ", self.name)
        start = 0
        end = recording.processor.sample_secs
        results = {}
        for c in recording.processor.chunks:

            # make predictions and put together with labels
            predictions = self.predict(c)[0]

            labeled_predictions = list(zip(self.labels, predictions))

            # Sort by score and filter
            sorted_predictions = sorted(
                labeled_predictions, key=operator.itemgetter(1), reverse=True
            )

            # Filter by recording.minimum_confidence so not to needlessly store full array for each chunk.
            labeled_predictions_filtered = [
                i for i in sorted_predictions if i[1] >= recording.minimum_confidence
            ]

            # Store results
            results[(start, end)] = labeled_predictions_filtered

            # Increment start and end
            start += recording.processor.sample_secs - recording.processor.overlap
            end = start + recording.processor.sample_secs

            self.results = results



from pathlib import Path
import importlib.util
import os
import time
import yaml
import inspect


# custom exception to have some more control over what is raised
class TFModelException(Exception):
    pass


def get_method_docstring(cls, method_name: str) -> str:
    """
    get_method_docstring Retrieve the docstring of a class method.

    Args:
        method_name (_type_): Name of the method to retrieve the docstring from

    Returns:
        str: Method docstring
    """
    # Get the method object from the class
    method = getattr(cls, method_name, None)
    if method is None:
        return None

    # Get the docstring of the method
    docstring = inspect.getdoc(method)
    return docstring


def update_dict_leafs_recursive(base, update):
    """
    update_dict_leafs_recursive Merge recursively two arbitrarily nested dictionaries such that only those leaves of 'base' are upated with the content of 'update' for which the given path in 'update' fully exists in 'base'.

    This function assumes that nodes in 'base' are only replaced, and 'update' does not add new nodes.

    Args:
        base (dict): Base dictionary to update.
        update (dict): dictionary to update 'base' with.
    """
    # basic assumption: update is a sub-tree of base with unknown entry point.
    if isinstance(base, dict) and isinstance(update, dict):

        for kb, vb in base.items():
            if kb in update:
                # overlapping element branch found
                if isinstance(vb, dict) and isinstance(update[kb], dict):
                    # follow branch if possible
                    update_dict_leafs_recursive(vb, update[kb])
                else:
                    # assign if not
                    base[kb] = update[kb]
            else:
                update_dict_leafs_recursive(vb, update)  # find entrypoint
    else:
        pass  # not found and no dictionaries - pass


def read_yaml(path: str):
    print(f"...reading config from {path}")
    """
        read_yaml Read the yaml basic config file for faunanet from path.
                It contains the install directory, data directory and other things used
                by faunanet internally.

        Args:
            path (str): Path to the yaml base config.

        Returns:
            dict: read base config file.
        """

    if Path(path).exists() is False:
        raise FileNotFoundError(f"The folder {path} does not exist")

    with open(Path(path)) as file:
        base_cfg = yaml.safe_load(file)

    return base_cfg


def load_model_from_file_tflite(path: str, num_threads: int = 1):
    """
    load_model_from_file_tflite Load model from a .tflite file.

    Args:
        path (str): Path to a .tflite model file to load
        num_threads (int, optional): Number of threads to use. Defaults to 1.

    Raises:
        FileNotFoundError: When the path given does not lead to an existing file
        TFModelException: When something goes wrong within tensorflow lite when loading the model or allocating tensors

    Returns:
        TensorflowLite interpreter: The loaded model
    """
    try:
        import tflite_runtime.interpreter as tflite
    except Exception:
        from tensorflow import lite as tflite

    if Path(path).exists() is False:
        raise FileNotFoundError(f"The desired model file does not exist: {path}")

    try:
        interpreter = tflite.Interpreter(model_path=path, num_threads=num_threads)
        return interpreter
    except Exception as e:
        raise TFModelException(e)


def load_model_from_file_pb(path: str, _):
    """
    load_model_from_file_pb Load a tensorflow model saved as .pb in tensorflow's saved_model format from file. The file to be loaded must be named 'saved_model.pb'

    Args:
        path (str): Path to a floder containing 'saved_model.pb'

    Raises:
        FileNotFoundError: When the given path does not lead to a folder with a 'saved_model.pb' file in it.
        TFModelException: When something goes wrong inside tensorflow when loading the model.

    Returns:
        Tensorflow model: The loaded model
    """
    import tensorflow as tf

    if "." in Path(path).name or ".pb" in Path(path).name:
        # tensorflow assumes a model file to be named "saved_model.pb" and the path given to be a directory
        path = Path(path).parent

    if Path(path).exists() is False:
        raise FileNotFoundError("The desired model file does not exist")

    try:
        model = tf.saved_model.load(path)
        return model
    except Exception as e:
        raise TFModelException(e)


def load_module(module_name: str, file_path: str):
    """
    load_module Load a python module from 'path' with alias 'alias'

    Args:
        module_name (str): module alias.
        file_path (str): Path to load the module from

    Returns:
        module: Python module that has been loaded
    """
    try:
        spec = importlib.util.spec_from_file_location(module_name, file_path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
    except Exception as e:
        raise RuntimeError(f"Error in loading module {file_path}") from e
    return module


def load_name_from_module(module_name: str, file_path: str, name: str):
    """
    load_name_from_module Load a python module from 'path' with alias 'alias'

    Args:
        module_name (str): module alias.
        file_path (str): Path to load the module from
        name (str): name to import
    Returns:
        module: Python module that has been loaded
    """
    module = load_module(module_name, file_path)
    return getattr(module, name)


def wait_for_file_completion(file_path: str, polling_interval=1) -> bool:
    """
    wait_for_file_completion Wait for a file to be fully written by checking when its size does no longer change.

    Args:
        file_path (str): filepath to check
        polling_interval (int, optional): how often file size should be checked. Defaults to 1.

    Returns:
        bool: True when the file size does not change any longer. False otherwise
    """
    initial_size = os.path.getsize(file_path)
    is_complete = False
    while True:

        time.sleep(polling_interval)

        current_size = os.path.getsize(file_path)

        if current_size == initial_size:
            is_complete = True
            break
        else:
            initial_size = current_size
    return is_complete



from faunanet import Recording
from faunanet import SpeciesPredictorBase
import faunanet.utils as utils

from pathlib import Path
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from time import sleep
from datetime import datetime
from copy import deepcopy
import yaml
import multiprocessing
import warnings
import csv
from contextlib import contextmanager
import traceback


class AnalysisEventHandler(FileSystemEventHandler):
    """
    AnalysisEventHandler Custom event handler that calls a function whenever
    the a new file with a matching extension is created in the watched directory and that waits for a
    multiprocessing.Event object to be true.

    Base:
        FileSystemEventHandler: watchdog.FilesystemEventHandler

    Methods:
    --------
    on_created(event): Run the callback whenever a newly created file with a matching pattern is found.
    """

    def __init__(
        self,
        watcher,
    ):
        """
        __init__ Create a new AnalysisEventHandler object.

        Args:
            callback (callable): Callback functionw hen
            pattern (str, optional): file apttern. Defaults to ".wav".
            watcher: (Watcher): Watcher this Handler is used with
        """
        self.pattern = watcher.pattern
        self.recording = watcher._set_up_recording(
            watcher.model_name,
            watcher.recording_config,
            watcher.species_predictor_config,
            watcher.model_config,
            watcher.preprocessor_config,
        )
        self.callback = watcher.analyze

    def on_created(self, event):
        """
        on_created Call the callback function of the caller object whenever a new file is created
                    that has the desired file extension. Waits on the stored multiprocessing.event and only
                    runs if its flag is true.

        Args:
            event (threading.Event): Event triggering the analysis, i.e., a new audio file appears and is ready to be processessed in the watched folder
        """
        if (
            Path(event.src_path).is_file()
            and Path(event.src_path).suffix == self.pattern
        ):
            self.callback(event.src_path, self.recording)


def watchertask(watcher):
    """
    watchertask Function to run in the watcher process.
    Checks if there is a new file every 'watcher.check_time' seconds and
    analyze the file with a sound classifier model.

    Args:
        watcher (Watcher): Watcher class to run this task with

    Raises:
        RuntimeError: When something goes wrong inside the analyzer process.
    """

    # build the recorder
    try:
        observer = Observer()

        event_handler = AnalysisEventHandler(
            watcher,
        )
        observer.schedule(event_handler, watcher.input, recursive=True)
        observer.start()
    except Exception as e:
        tb = traceback.format_exc()
        watcher.exception_queue.put((e, tb))

    try:
        while True:
            sleep(watcher.check_time)
    except KeyboardInterrupt:
        observer.stop()
    except Exception as e:
        observer.stop()
        tb = traceback.format_exc()
        watcher.exception_queue.put((e, tb))

    observer.join()


class Watcher:
    """
    Class that watches a directory and applies a classifier model to each new file that is created in it.
    Supports model exchange on the fly.
    To this end, it creates a daemon thread in which the classifier runs,
    such that commands can be processed in the main process.

    Methods:
    --------
    change_analyzer: Exchange the classifier model with one that is present in the model dir.
    analyze: Analyze a file and write the results to the output directory.
    save_results: Save the results to the output directory as a csv file.
    stop: Pause the watcher thread.
    go_on: Continue the watcher thread
    check_analysis: Go over the input and output directory and repeat analysis for all files for which there is not corresponding file in the output directory.
    """

    def _write_config(
        self,
    ):
        """
        _write_config Write out the set of parameters a Watcher instance has been created with.

        """
        config = {
            "Analysis": {
                "input": str(self.input),
                "output": str(self.output),
                "delete_recordings": self.delete_recordings,
                "pattern": self.pattern,
                "model_name": self.model_name,
                "check_time": self.check_time,
                "model_dir": str(self.model_dir),
                "Preprocessor": deepcopy(self.preprocessor_config),
                "Model": deepcopy(self.model_config),
                "Recording": deepcopy(self.recording_config),
                "SpeciesPredictor": deepcopy(self.species_predictor_config),
            }
        }

        try:
            with open(self.output / "config.yml", "w") as ymlfile:
                yaml.safe_dump(config, ymlfile)
        except Exception as e:
            raise RuntimeError("Error during config writing ", e) from e

    def _set_up_recording(
        self,
        model_name: str,
        recording_config: dict,
        species_predictor_config: dict,
        model_config: dict,
        preprocessor_config: dict,
    ) -> Recording:
        """
        _set_up_recording Build a new recording from configs

        Args:
            model_name (str): name of the model to use
            recording_config (dict): recording config
            species_predictor_config (dict): species_predictor config
            model_config (dict): model config data
            preprocessor_config (dict): preprocessor config data

        Raises:
            ValueError: In case the species presence predictor is used. When the an error occurs during the creation of the species predictor model.

        Returns:
            Recording: New instance of SpeciesRecording created with config dictionaries held by the caller.
        """
        recording_config = deepcopy(recording_config)

        preprocessor = utils.load_name_from_module(
            "pp",
            self.model_dir / Path(model_name) / "preprocessor.py",
            "Preprocessor",
        )(**preprocessor_config)

        model = utils.load_name_from_module(
            "mo", self.model_dir / Path(model_name) / "model.py", "Model"
        )(model_path=self.model_dir / model_name, **model_config)

        # process species range predictor
        if all(name in recording_config for name in ["date", "lat", "lon"]) and all(
            recording_config[name] is not None for name in ["date", "lat", "lon"]
        ):

            try:
                # we can use the species predictor
                species_predictor = SpeciesPredictorBase(
                    model_path=self.model_dir / model_name,
                    **species_predictor_config,
                )

                recording_config["species_predictor"] = species_predictor

            except Exception as e:
                raise ValueError(
                    "An error occured during species range predictor creation. Does you model provide a model file called 'species_presence_model'?"
                ) from e

        # make sure the date is set correctly
        if "date" in recording_config:
            if recording_config["date"] == "today":
                recording_config["date"] = datetime.now()
            elif isinstance(recording_config["date"], str):
                recording_config["date"] = datetime.strptime(
                    recording_config["date"], "%d/%m/%Y"
                )
            else:
                # nothing to do here
                pass

        # create recording object
        # species predictor is applied here once and then used for all the analysis calls that may follow
        return Recording(preprocessor, model, "", **recording_config)

    def _restore_old_state(self, old_state: dict):
        """
        _restore_old_state Restore state of the watcher instance to the state it had before a change was made.

        Args:
            old_state (dict):  Dictionary containing the old state of the calling object
        """
        self.model_name = old_state["model_name"]
        self.preprocessor_config = old_state["preprocessor_config"]
        self.model_config = old_state["model_config"]
        self.recording_config = old_state["recording_config"]
        self.species_predictor_config = old_state["species_predictor_config"]
        self.pattern = old_state["pattern"]
        self.check_time = old_state["check_time"]
        self.delete_recordings = old_state["delete_recordings"]

    @contextmanager
    def _backup_and_restore_state(self):
        """
        _backup_and_restore_state Helper context manager that creates a backup of the current object state and restores it in case of an error.

        Raises:
            RuntimeError When an error occurs during the restart of the watcher process.

        Yields:
            dict: Dictionary containing the old state of the calling object
        """
        old_state = {
            "model_name": deepcopy(self.model_name),
            "preprocessor_config": deepcopy(self.preprocessor_config),
            "model_config": deepcopy(self.model_config),
            "recording_config": deepcopy(self.recording_config),
            "species_predictor_config": deepcopy(self.species_predictor_config),
            "pattern": deepcopy(self.pattern),
            "check_time": self.check_time,
            "delete_recordings": self.delete_recordings,
        }

        try:
            yield old_state
        except Exception as e:
            self._restore_old_state(old_state)

            if self.is_running:
                self.stop()

            raise RuntimeError(
                "Error when while trying to change the watcher process, any changes made have been undone. The process needs to be restarted manually. This operation may have led to data loss."
            ) from e

    def __init__(
        self,
        indir: str,
        outdir: str,
        model_dir: str,
        model_name: str,
        preprocessor_config: dict = None,
        model_config: dict = None,
        recording_config: dict = None,
        species_predictor_config: dict = None,
        pattern: str = ".wav",
        check_time: int = 1,
        delete_recordings: str = "never",
    ):
        """
        __init__ Create a new Watcher object.

        Args:
            indir (str): Directory to read files from
            outdir (str): Directory to write the analysis results to.
                        A new directory with a timestamp for name will be
                        created within it to store the result files in.
            model_dir (str): Directory where downloaded models are stored
            model_name (str): name of the model to use. Must match the directory name it is stored in, e.g., 'birdnet_default'.
            preprocessor_config (dict, optional): Keyword arguments for the preprocessor of the model. Defaults to {}.
            model_config (dict, optional): Keyword arguments for the model instance. Defaults to {}.
            recording_config (dict, optional): Keyword arguments for the internal Recording. Defaults to {}.
            species_predictor_config (dict, optional): Keyword arguments for a species presence predictor model. Defaults to {}.
            pattern (str, optional): filename pattern to look for. defaults to '.wav'.
            check_time(int, optional): Sleep time of the watcher between checks for new files in seconds. Defaults to 1.
            delete_recordings(str, optional): Mode for data clean up. Can be one of "never" or "always".
                                            "never" keeps recordings around indefinitely. 'always' deletes the recording
                                            immediatelly after analysis. Defaults to 'never'.
        Raises:
            ValueError: When the indir parameter is not an existing directory.
            ValueError: When the outdir parameter is not an existing directory.
            ValueError: When the model_dir parameter is not an existing directory.
            ValueError: When the model name does not correspond to a directory in the 'model_dir' directory in which available models are stored.
        """
        if preprocessor_config is None:
            preprocessor_config = {}

        if model_config is None:
            model_config = {}

        if recording_config is None:
            recording_config = {}

        if species_predictor_config is None:
            species_predictor_config = {}

        # set up data to use
        self.input = Path(indir)

        if self.input.is_dir() is False:
            raise ValueError("Input directory does not exist")

        self.outdir = Path(outdir)

        if self.outdir.is_dir() is False:
            raise ValueError("Output directory does not exist")

        self.output = None
        self.old_output = None

        self.model_dir = Path(model_dir)

        if self.model_dir.is_dir() is False:
            raise ValueError("Model directory does not exist")

        if (self.model_dir / model_name).is_dir() is False:
            raise ValueError("Given model name does not exist in model directory")

        self.pattern = pattern

        self.model_name = model_name

        self.watcher_process = None

        self.exception_queue = multiprocessing.Queue()

        self.may_do_work = multiprocessing.Event()

        self.is_done_analyzing = multiprocessing.Event()

        self.check_time = check_time

        if delete_recordings not in ["never", "always"]:
            raise ValueError("'delete_recordings' must be in 'never', 'always'")

        self.delete_recordings = delete_recordings

        self.first_analyzed = multiprocessing.Value("i", 0)

        self.last_analyzed = multiprocessing.Value("i", 0)

        self.preprocessor_config = deepcopy(preprocessor_config)

        self.model_config = deepcopy(model_config)

        self.recording_config = deepcopy(recording_config)

        self.species_predictor_config = deepcopy(species_predictor_config)

        self.batchfile_name = "batch_info.yml"

    @property
    def output_directory(self):
        return str(self.output)

    @property
    def input_directory(self):
        return str(self.input)

    @property
    def is_running(self):
        if self.watcher_process is not None:
            return self.watcher_process.is_alive()
        else:
            return False

    @property
    def is_sleeping(self):
        if self.watcher_process is not None:
            return self.may_do_work.is_set() is False
        else:
            return False

    def analyze(self, filename: str, recording: Recording):
        """
        analyze Analyze a file pointed to by 'filename' and save the results as csv file to 'output'.

        Args:
            filename (str): path to the file to analyze.
            recording (Recording): recording object to use
        """
        self.may_do_work.wait()  # wait until parent process allows the worker to pick up work

        recording.path = filename

        recording.analyzed = False  # reactivate

        # make the main process wait on the finish signal to make sure no
        # corrupt files are produced
        self.is_done_analyzing.clear()

        self.last_analyzed.value = int(Path(filename).stat().st_ctime)

        if self.first_analyzed.value == 0:
            self.first_analyzed.value = self.last_analyzed.value

        recording.analyze()

        results = recording.detections

        self.save_results(self.output, results, suffix=Path(filename).stem)

        self.is_done_analyzing.set()  # give good-to-go for main process

        if self.delete_recordings == "always":
            Path(filename).unlink()

    def save_results(self, outfolder: str, results: list, suffix=""):
        """
        save_results Save results to csv file.

        Args:
            outfolder (str): folder to save the results in
            suffix (str, optional): _description_. Defaults to "".
        """

        with open(Path(outfolder) / Path(f"results_{suffix}.csv"), "w") as csvfile:
            if len(results) == 0:
                writer = csv.writer(csvfile)
                writer.writerow([])
            else:
                colnames = results[0].keys()
                writer = csv.DictWriter(csvfile, fieldnames=colnames)
                writer.writeheader()
                writer.writerows(results)

    def start(self):
        """
        start Watch the directory the caller has been created with and analyze all newly created files matching a certain file ending. \
            Creates a new daemon process in which the analysis function runs.

        Raises:
            RuntimeError: When the watcher process is running already.
        """
        if self.is_running:
            raise RuntimeError("watcher process still running, stop first.")

        self.old_output = self.output
        self.output = Path(self.outdir) / Path(datetime.now().strftime("%y%m%d_%H%M%S"))
        self.output.mkdir(exist_ok=True, parents=True)
        self.first_analyzed.value = 0
        self.last_analyzed.value = 0

        self._write_config()

        try:
            print("start the watcher process")
            # create a background watchertask such that the command is handed back to the parent process
            self.watcher_process = multiprocessing.Process(
                target=watchertask, args=(self,)
            )
            self.watcher_process.daemon = True
            self.watcher_process.name = "watcher_process"
            self.watcher_process.start()
            self.may_do_work.set()
            self.is_done_analyzing.clear()

        except Exception as e:

            if self.output.is_dir():
                for filename in self.output.iterdir():
                    filename.unlink()
                self.output.rmdir()

            self.may_do_work.clear()
            self.is_done_analyzing.clear()
            self.watcher_process = None

            raise RuntimeError(
                "Something went wrong when starting the watcher process, undoing changes and returning"
            ) from e

    def restart(self):
        """
        restart Restart the watcher process. Must be called when, e.g., new models have been loaded or the input or output has changed.
        """
        print("trying to restart the watcher process")
        self.stop()
        self.start()

    def pause(self):
        """
        pause Pause the watcher process.

        Raises:
            RuntimeError: When the watcher process is not running.
        """
        if self.watcher_process is not None and self.watcher_process.is_alive():
            is_flag_set = self.is_done_analyzing.wait(
                timeout=30
            )  # wait for the analysis finish event

            if is_flag_set is False:
                warnings.warn(
                    "pause timeout expired, setting on pause without regard for status"
                )

            self.may_do_work.clear()
        else:
            raise RuntimeError("Cannot pause watcher process, is not alive anymore.")

    def go_on(self):
        """
        go_on Continue the watcher thread.

        Raises:
            RuntimeError: When the watcher process is not running anymore
        """
        if self.watcher_process is not None and self.watcher_process.is_alive():
            print("continue the watcher process")
            self.may_do_work.set()
        else:
            raise RuntimeError("Cannot continue watcher process, is not alive anymore.")

    def stop(self):
        """
        stop Stop the watcher process if it is still running.

        Raises:
            RuntimeError: When the watcher process is not running anymore
        """
        if self.watcher_process is not None and self.watcher_process.is_alive():
            print("trying to stop the watcher process")
            flag_set = self.is_done_analyzing.wait(timeout=30)

            if flag_set is False:
                warnings.warn("stop timeout expired, terminating watcher process now.")

            try:
                self.watcher_process.terminate()
                self.watcher_process.join()
                self.watcher_process.close()
                self.watcher_process = None
                self.may_do_work.clear()
                self.is_done_analyzing.set()
            except Exception as e:
                raise RuntimeError(
                    "Something went wrong when trying to stop the watcher process"
                ) from e

            with open(self.output / self.batchfile_name, "w") as batch_info:
                yaml.safe_dump(
                    {
                        "first": self.first_analyzed.value,
                        "last": self.last_analyzed.value,
                    },
                    batch_info,
                )
        else:
            raise RuntimeError("Cannot stop watcher process, is not alive anymore.")

    def change_analyzer(
        self,
        model_name: str,
        preprocessor_config: dict = None,
        model_config: dict = None,
        recording_config: dict = None,
        species_predictor_config: dict = None,
        pattern: str = ".wav",
        check_time: int = 1,
        delete_recordings: str = "never",
    ):
        """
        change_analyzer Change classifier model to the one indicated by name.
        The given model name must correspond to the name of a folder in the
        faunanet models directory created upon install. A clean-up method is
        run after model change to compensate any loss of analysis data that
        may occur during the restart of the watcher process with a different
        model.

        Args:
            model_name (str): Name of the model to be used
            preprocessor_config (dict, optional): Parameters for preprocessor given as key(str): value. If empty, default parameters of the preprocessor will be used. Defaults to {}.
            model_config (dict, optional): Parameters for the model given as key(str): value. If empty, default parameters of the model will be used. Defaults to {}.
            recording_config (dict, optional): Parameters for the underlyin Recording object. If empty, default parameters of the recording will be used. Defaults to {}.
            species_predictor_config (dict, optional): _description_. If empty, default parameters of the species predictor will be used. Defaults to {}.
            Make sure the model you use is compatible with a species predictor before supplying these.
            pattern (str, optional): filename pattern to look for. defaults to '.wav'.
            check_time(int, optional): Sleep time of the watcher between checks for new files in seconds. Defaults to 1.
            delete_recordings(str, optional): Mode for data clean up. Can be one of "never" or "always".
                                            "never" keeps recordings around indefinitely. 'always' deletes the recording
                                            immediatelly after analysis. Defaults to 'never'.

        Raises:
            ValueError: _description_
            RuntimeError: _description_
        """

        if self.watcher_process is None or self.is_running is False:
            raise RuntimeError("Watcher not running, cannot change analyzer")

        # import and build new model, pause the analyzer process,
        # change the model, resume the analyzer
        if preprocessor_config is None:
            preprocessor_config = {}

        if model_config is None:
            model_config = {}

        if recording_config is None:
            recording_config = {}

        if species_predictor_config is None:
            species_predictor_config = {}

        if (self.model_dir / model_name).is_dir() is False:
            raise ValueError("Given model name does not exist in model dir.")

        with self._backup_and_restore_state() as old_state:
            self.model_name = model_name
            self.preprocessor_config = preprocessor_config
            self.model_config = model_config
            self.recording_config = recording_config
            self.species_predictor_config = species_predictor_config
            self.pattern = pattern
            self.check_time = check_time
            self.delete_recordings = delete_recordings

            self.restart()

            try:
                i = 0
                while self.last_analyzed.value <= self.first_analyzed.value:
                    sleep(3)
                    if i > 10:
                        break
                    i += 1
                warnings.warn(
                    "No data was incoming in 30s, trying to clean up data anyway."
                )
                self.clean_up()

            except Exception as e:
                status = (
                    "still running"
                    if self.is_running
                    else "not running, any changes have been undone."
                )

                cause = e.__cause__ if e.__cause__ is not None else e

                if self.is_running is False:
                    self.restore_old_state(old_state)
                    self.may_do_work.clear()
                    self.is_done_analyzing.set()
                    self.output = Path(self.outdir) / Path(
                        datetime.now().strftime("%y%m%d_%H%M%S")
                    )
                raise RuntimeError(
                    f"Error when cleaning up data after analyzer change, watcher is {status}. The cause was {cause}. This error may have lead to corrupt data in newly created analysis files."
                ) from e

    def _get_clean_up_limits(self, older_output: str, newer_output: str) -> tuple:
        """
        _get_clean_up_limits Determine the lower and upper time limit for the clean up of the older output directory.

        Args:
            older_output (str): Older output folder that is actually cleaned up.
            newer_output (str): Newer output folder that is used to determine the upper time limit for the clean up.

        Raises:
            RuntimeError: When the newer output directory is not the currently worked on, but also has no batch info file.

        Returns:
            tuple: (lower time limit, upper time limit) pair of timestamps to use for the clean up.
        """
        with open(older_output / self.batchfile_name, "r") as batch_info:
            lower_limit = yaml.safe_load(batch_info)["last"]

        if newer_output is None:
            upper_limit = int(datetime.now().timestamp())

        elif Path(newer_output / self.batchfile_name).is_file():
            with open(newer_output / self.batchfile_name, "r") as batch_info:
                upper_limit = yaml.safe_load(batch_info)["first"]

        elif newer_output != self.output:
            raise RuntimeError(
                f"{newer_output} is not the current output directory {self.output} but has no batchinfo.yml file."
            )

        else:
            upper_limit = self.first_analyzed.value
        return lower_limit, upper_limit

    def _clean_up_between(self, older_output: str, newer_output: str):
        """
        _clean_up_between Run clean up on the older output directory between the last analyzed file and the first analyzed file in the newer output directory.
                          This function assumes that older_output and newer_output are consecutive output directories of the watcher process. Passing it
                          non-consecutive directories is undefined and may lead to unexpected results.
                          If newer_output is None, the current time is used as upper limit and all files in the older output directory are analyzed that have
                          not been analyzed by the current process.

        Args:
            older_output (str): Older output directory. This is done one that gets cleaned up.
            newer_output (str): Newer output directory. This is used to determine the upper limit time for the clean up.
        """
        if older_output == self.output and self.is_running:
            warnings.warn(
                "Cannot clean up current output directory while watcher is running"
            )
            return

        with open(older_output / "config.yml", "r") as ymlfile:
            cfg = yaml.safe_load(ymlfile)

        lower_limit, upper_limit = self._get_clean_up_limits(older_output, newer_output)

        input_folder = Path(cfg["Analysis"]["input"])
        outputfolder = cfg["Analysis"]["output"]

        audiofiles = [
            f
            for f in input_folder.iterdir()
            if f.suffix == cfg["Analysis"]["pattern"]
            and not Path(outputfolder, f"results_{f.stem}.csv").is_file()
            and lower_limit < f.stat().st_ctime < upper_limit
        ]

        if len(audiofiles) > 0:
            recording = self._set_up_recording(
                cfg["Analysis"]["model_name"],
                cfg["Analysis"]["Recording"],
                cfg["Analysis"]["SpeciesPredictor"],
                cfg["Analysis"]["Model"],
                cfg["Analysis"]["Preprocessor"],
            )

            for audiofile in audiofiles:
                recording.path = audiofile
                recording.analyzed = False
                recording.analyze()
                results = recording.detections
                self.save_results(outputfolder, results, suffix=str(audiofile.stem))

                if cfg["Analysis"]["delete_recordings"] == "always":
                    audiofile.unlink()

            with open(older_output / "missings.txt", "w") as missing:
                for audiofile in audiofiles:
                    missing.write(f"{audiofile}\n")

    def clean_up(self):
        """
        clean_up Run cleanup on the all the output directories of the watcher process that reside in the current output base directory.

        """

        folders = sorted(
            filter(lambda f: f.is_file() is False, list(self.outdir.iterdir())),
            key=lambda x: x.stat().st_ctime,
        )

        folders.append(None)  # add dummy to include the last folder

        if folders == [
            None,
        ]:
            raise RuntimeError("No output folders found to clean up")
        else:
            for i in range(1, len(folders)):
                self._clean_up_between(folders[i - 1], folders[i])



from pathlib import Path
from platformdirs import user_config_dir, user_cache_dir
import cmd
import time
import traceback

from faunanet import Watcher
import faunanet.faunanet_setup as sps
from faunanet.utils import read_yaml, update_dict_leafs_recursive, get_method_docstring


def process_line_into_kwargs(line: str, keywords: list = None) -> dict:
    """
    process_line_into_kwargs Process string into keyword argument dictionary

    Args:
        line (str): raw string input that contains keyword arguments
        keywords (list, optional): list of kwargs to look for in the input line. Defaults to None.

    Raises:
        ValueError: When the input does not adhere to the --name=<arg> structure
        ValueError: When the keywords are not provided in the input line
        ValueError: When an expected keyword is not found in the input line

    Returns:
        dcit: dictionary with keyword: value pairs, where the values are strings
    """
    if line == "":
        return {}

    if "=" not in line:
        raise ValueError("Invalid input. Expected options structure is --name=<arg>")

    if keywords is None or len(keywords) == 0:
        raise ValueError("Keywords must be provided with passed line")

    for k in keywords:
        if k not in line:
            raise ValueError(f"Keyword {k} not found in passed line")

    kwargs = {
        name.lstrip("-"): key
        for part in line.split(" ")
        if "=" in part
        for name, key in [part.split("=")]
    }

    return kwargs


class FaunanetCmd(cmd.Cmd):
    """
    FaunanetCmd is a command-line interface for interacting with a FaunanetWatcher instance.

    This class provides a set of commands for setting up, starting, stopping, and managing a watcher process that analyzes incoming files in a directory. It also provides commands for getting the status of the watcher process and the current setup of faunanet.

    Attributes:
        prompt (str): The prompt string displayed in the command-line interface.
        intro (str): The introductory message displayed when the command-line interface starts.
        watcher: The watcher process that is being managed.
        running (bool): Indicates whether the command-line interface is currently running.
    """

    prompt = "(faunanet)"
    intro = "Welcome to faunanet! Type help or ? to list commands.\n"

    def __init__(self):
        """
        __init__ Create a new FaunanetCmd instance.

        """
        super().__init__()
        self.watcher = None
        self.running = True

    def update_parameters(self, cfgpath: str = None):
        """
        update_parameters Update watcher parameter dictionaries read from yaml files.
                        A default configuration file is used if no custom configuration file is provided.
                        Otherwise, the default config for the desired model is updated with the custom config.

        Args:
            cfgpath (str, optional): Path to a config file. Defaults to None.

        Returns:
            _type_: _description_
        """
        folders = read_yaml(Path(user_config_dir()) / Path("faunanet") / "install.yml")[
            "Directories"
        ]

        model_dir = Path(folders["models"]).expanduser()

        if cfgpath is not None:
            custom_config = read_yaml(Path(cfgpath).expanduser())
        else:
            custom_config = read_yaml(
                model_dir / Path("birdnet_default") / "default.yml"
            )

        fallback_modelname = (
            self.watcher.model_name if self.watcher is not None else "birdnet_default"
        )

        modelname = (
            Path(custom_config["Analysis"]["modelname"])
            if "modelname" in custom_config["Analysis"]
            else fallback_modelname
        )

        config = read_yaml(model_dir / modelname / "default.yml")

        update_dict_leafs_recursive(config, custom_config)

        return config

    def process_arguments(
        self,
        line: str,
        keywords: list,
        do_no_inputs: callable = lambda s: None,
        do_with_inputs: callable = lambda s: None,
        do_with_failure: callable = lambda s, args, e: None,
    ) -> tuple:
        """
        process_arguments Process the argument string of a command into its individual parts, based on a '--name=<arg>' structure. Depending on the input, different actions are taken as given in the arguments.
        Args:
            line (str): Argument string
            keywords (list): expected argument names in the form of '--name=<arg>'
            do_no_inputs (callable, optional): Function to call when no inputs are given. Defaults to lambda s:None (do nothing).
            do_with_inputs (callable, optional): Function to call with the correct input structure. Defaults to lambda s:None.
            do_with_failure (callable, optional): Function to call when an exception is thrown during processing. Defaults to lambda s, e: None.

        Returns:
            tuple: _description_
        """
        inputs = None
        try:
            inputs = process_line_into_kwargs(line, keywords)
        except Exception as e:
            do_with_failure(self, inputs, e)
            return inputs, True

        if len(inputs) > len(keywords):
            print(
                f"Invalid input. Expected {len(keywords)} blocks of the form --name=<arg> with names {keywords}"
            )
            return inputs, True
        elif len(inputs) == 0:
            try:
                do_no_inputs(self, inputs)
                return inputs, False
            except Exception as e:
                do_with_failure(self, inputs, e)
                return inputs, True
        else:
            try:
                do_with_inputs(self, inputs)
                return inputs, False
            except Exception as e:
                do_with_failure(self, inputs, e)
                return inputs, True

    def dispatch_on_watcher(
        self,
        do_is_none: callable = lambda s: None,
        do_is_sleeping: callable = lambda s: None,
        do_is_running: callable = lambda s: None,
        do_else: callable = lambda s: None,
        do_failure: callable = lambda s, e: None,
    ):
        """
        dispatch_on_watcher Process different watcher states, using different callables. Useful for error catching and handling when a command should be issued to a watcher.

        Args:
            do_is_none (callable, optional): Function to call when there is no watcher. Defaults to lambda s:None. (do nothing)
            do_is_sleeping (callable, optional): Function to call when the watcher is sleeping. Defaults to lambda s:None.
            do_is_running (callable, optional): Function to call when the watcher is running. Defaults to lambda s:None.
            do_else (callable, optional): Function to call when the watcher has been stopped or has been created but is not running. Defaults to lambda s:None.
            do_failure (callable, optional): Function to call upon an exception occuring in any of the above. Defaults to lambda s, e: None (do nothing). Make sure to override this to get proper error handling.
        """
        if self.watcher is None:
            try:
                do_is_none(self)
            except Exception as e:
                do_failure(self, e)
        elif self.watcher.is_sleeping:
            try:
                do_is_sleeping(self)
            except Exception as e:
                do_failure(self, e)
        elif self.watcher.is_running:
            try:
                do_is_running(self)
            except Exception as e:
                do_failure(self, e)
        else:
            try:
                do_else(self)
            except Exception as e:
                do_failure(self, e)

    def print_error(self, message: str):
        """
        print_error Print error message and traceback. Used for providing information about exceptions.

        Args:
            message (str): Error message to print
        """
        print(
            message,
            flush=True,
        )
        print("Traceback: ")
        traceback.print_exc()

    def handle_exit(self):
        """
        handle_exit  Stops the watcher if it is running upon exiting the shell if it is running.
        """
        self.running = False
        if self.watcher is not None and self.watcher.is_running:
            try:  # Try to stop the watcher
                self.watcher.stop()
                self.wait_for_watcher_event(
                    lambda s: s.watcher.is_running is False, limit=20, waiting_time=3
                )
            except Exception as e:
                self.print_error(
                    f"Could not stop watcher: {e} caused by {e.__cause__}. Watcher process will be killed now."
                )

                if self.watcher.is_running:
                    self.watcher.watcher_process.kill()

    def wait_for_watcher_event(
        self, event: callable, limit: int = 10, waiting_time: int = 5
    ):
        """
        wait_for_watcher_event Wait for  defined event happening to the watcher process, and block the caller until then.

        Args:
            event (callable): Event to wait for. Must be a function that takes the current instance of the shell as an argument and returns a boolean.
            limit (int, optional): Timeout limit for the waiting. Timeout in seconds is limit * waiting_time. Defaults to 10.
            waiting_time (int, optional): Waiting time for each test. Defaults to 5.
        """
        i = 0
        while True:
            if event(self):
                break
            elif i > limit:
                print(
                    "Error while performing watcher interaction: Timeout.", flush=True
                )
                break
            else:
                time.sleep(waiting_time)
                i += 1

    # Command definitions

    def do_help(self, line: str):
        """
        do_help Print help strings for available commands. The argument string is optional and can be the name of a command to get more info about.

        Args:
            line (str): name of a command to get more info about.
        """
        if line == "":
            print("Commands: ")
            print(
                "set_up: set up faunanet for usage. Usage: 'set_up --cfg=<path>' with <path> being a path to a custom yaml configuration file. When no argumnet is provided, the birdnet default is used."
            )
            print(
                "start: start a watcher for analyzing incoming files in a directory. Usage: 'start --cfg=<path>'. When no argumnet is provided, the default from birdnetlib is used."
            )
            print(
                "stop: stop a previously started watcher. This command is used without any further arguments"
            )
            print(
                "pause: pause a running watcher. This command is used without any further arguments"
            )
            print(
                "continue: continue a paused watcher. This command is used without any further arguments"
            )
            print(
                "restart: restart an existing watcher. This command is used without any further arguments"
            )
            print(
                "change analyzer: change the analyzer of a running watcher. Usage: 'change_analyzer --cfg=<path>' with <path> being a path to a custom yaml configuration file"
            )
            print(
                "cleanup: cleanup the output directory of the watcher, assuring data consistency. This command is used without any further arguments"
            )
            print(
                "status: get the current status of the watcher process. This command is used without any further arguments"
            )
            print(
                "get_setup_info: get information about the current setup of faunanet. This command is used without any further arguments"
            )
            print(
                "exit: leave this shell. This command is used without any further arguments"
            )
            print(
                "For further information on specific commands, type 'help <command>' or '? <command>'."
            )
        else:
            print(get_method_docstring(self, "do_" + line), flush=True)

    def do_set_up(self, line: str):
        """
        do_set_up Setup faunanet. This creates a `faunanet` folder in the user's home directory and copies the default configuration files into os_standard_config_dir/faunanet. If a custom configuration file is provided, it will be used instead of the default.

        Args:
            line (str): Optional path relative to the user's home directory to a custom configuration file. The file must be a yaml file with the same structure as the default configuration file.
        """

        def do_no_inputs(_, __):
            print("No config file provided, falling back to default\n")
            sps.set_up(None)

        self.process_arguments(
            line,
            ["--cfg"],
            do_no_inputs=lambda self, inputs: do_no_inputs(self, inputs),
            do_with_inputs=lambda self, inputs: sps.set_up(
                Path(inputs["cfg"]).expanduser().resolve()
            ),
            do_with_failure=lambda self, inputs, e: print(
                "Could not set up faunanet", e, "caused by: ", e.__cause__
            ),
        )

    def do_start(self, line: str):
        """
        do_start Start a new faunanet watcher process. Only can be started if no other watcher is currently running.

        Args:
            line (str): optional argument --cfg=<path> to provide a custom configuration file.
        """

        # helper closures to use with self.dispatch_on_watcher. No exception handling is necesssary in the functions below,
        # since this is handled by the dispatch_on_watcher method.
        def build_watcher(cfg: dict):
            self.watcher = Watcher(
                indir=Path(cfg["Data"]["input"]).expanduser().resolve(),
                outdir=Path(cfg["Data"]["output"]).expanduser().resolve(),
                model_dir=Path(cfg["Analysis"]["model_dir"]).expanduser().resolve(),
                model_name=cfg["Analysis"]["modelname"],
                model_config=cfg["Analysis"]["Model"],
                preprocessor_config=cfg["Data"]["Preprocessor"],
                recording_config=cfg["Analysis"]["Recording"],
                species_predictor_config=cfg["Analysis"].get("SpeciesPredictor", None),
                pattern=cfg["Analysis"]["pattern"],
                check_time=cfg["Analysis"]["check_time"],
                delete_recordings=cfg["Analysis"]["delete_recordings"],
            )

        def start_watcher():
            self.watcher.start()
            self.wait_for_watcher_event(
                lambda s: s.watcher.is_running, limit=20, waiting_time=3
            )

        # actual build and start process
        if Path(user_config_dir(), "faunanet").exists() is False:
            print("No installation found - please run the setup command first")
            return

        args, error = self.process_arguments(
            line,
            ["--cfg"],
            do_no_inputs=lambda _, __: print(
                "No config file provided, falling back to default"
            ),
            do_with_inputs=lambda _, __: None,
            do_with_failure=lambda _, __, e: print(
                "Something in the start command parsing went wrong. Check your passed commands. Caused by: ",
                e,
                flush=True,
            ),
        )

        if error is True:
            print("An error occurred while processing arguments.")
            print("traceback: ")
            traceback.print_exc()
            return

        cfgpath = Path(args["cfg"]).expanduser().resolve() if "cfg" in args else None

        cfg = None
        try:
            cfg = self.update_parameters(cfgpath)
        except Exception as e:
            self.print_error(
                f"An error occured while trying to update the parameters: {e} caused by {e.__cause__}"
            )
            return

        self.dispatch_on_watcher(
            do_is_none=lambda _: build_watcher(cfg),
            do_is_sleeping=lambda _: lambda _: print(
                "The watcher is running. Cannot be started again with different parameters. Try 'change_analyzer' to use different parameters.",
                flush=True,
            ),
            do_is_running=lambda _: print(
                "The watcher is running. Cannot be started again with different parameters. Try 'change_analyzer' to use different parameters.",
                flush=True,
            ),
            do_else=lambda _: print(
                "It appears that there is a watcher process that is not running. Trying to start with current parameters. Use  the 'change_analyzer' command to change the parameters.",
                flush=True,
            ),
            do_failure=lambda _, e: self.print_error(
                f"An error occured while trying to build the watcher: {e} caused by {e.__cause__}"
            ),
        )

        self.dispatch_on_watcher(
            do_is_none=lambda _: print(
                "There seems to be no watcher. Cannot start.",
                flush=True,
            ),
            do_is_sleeping=lambda _: print(
                "The watcher is running. Cannot be started again with different parameters. Try 'change_analyzer' to use different parameters.",
                flush=True,
            ),
            do_is_running=lambda _: print(
                "The watcher is running. Cannot be started again with different parameters. Try 'change_analyzer' to use different parameters.",
                flush=True,
            ),
            do_else=lambda _: start_watcher(),
            do_failure=lambda _, e: self.print_error(
                f"Something went wrong while trying to start the watcher: {e} caused by  {e.__cause__}. A new start attempt can be made when the error has been addressed."
            ),
        )

    def do_stop(self, line: str):
        """
        do_stop Stop a running faunanet watcher process.
        """
        if len(line) > 0:
            print("Invalid input. Expected no arguments.")
            return

        def handle_failure(s: cmd, e: Exception):
            self.print_error(
                f"Could not stop watcher: {e} caused by {e.__cause__}. Watcher process will be killed now and all resources released. This may have left data in a corrupt state. A new watcher must be started if this session is to be continued."
            )

            if s.watcher.is_running:
                s.watcher.watcher_process.kill()
            s.watcher = None

        self.dispatch_on_watcher(
            do_is_none=lambda _: print("Cannot stop watcher, no watcher present"),
            do_is_sleeping=lambda self: self.watcher.stop(),
            do_is_running=lambda self: self.watcher.stop(),
            do_else=lambda _: print("Cannot stop watcher, is not running"),
            do_failure=handle_failure,
        )

        if self.watcher is not None:
            self.wait_for_watcher_event(
                lambda s: s.watcher.is_running is False, limit=20, waiting_time=3
            )

    def do_cleanup(self, line: str):
        """
        do_cleanup Run the cleanup method of the watcher process to assure data consistency

        Args:
            line (str): Empty string, no arguments expected
        """
        if len(line) > 0:
            print("Invalid input. Expected no arguments.")
            return

        self.dispatch_on_watcher(
            do_is_none=lambda _: print(
                "Cannot run cleanup, no watcher present", flush=True
            ),
            do_is_sleeping=lambda self: self.watcher.clean_up(),
            do_is_running=lambda self: self.watcher.clean_up(),
            do_else=lambda self: self.watcher.clean_up(),
            do_failure=lambda s, e: self.print_error(
                f"Error while running cleanup: {e}"
            ),
        )

    def do_exit(self, line: str):
        """
        do_exit Leave the faunanet shell
        """
        self.running = False
        if len(line) > 0:
            print("Invalid input. Expected no arguments.")
            return

        print("Exiting faunanet shell")
        return True

    def do_pause(self, line: str):
        """
        do_pause Pause the watcher process.

        Args:
            line (str): Empty string, no arguments expected
        """
        if len(line) > 0:
            print("Invalid input. Expected no arguments.", flush=True)
            return

        self.dispatch_on_watcher(
            do_is_none=lambda _: print(
                "Cannot pause watcher, no watcher present", flush=True
            ),
            do_is_running=lambda _: self.watcher.pause(),
            do_is_sleeping=lambda _: print(
                "Cannot pause watcher, is already sleeping", flush=True
            ),
            do_else=lambda self: print(
                "Cannot pause watcher, is not running", flush=True
            ),
            do_failure=lambda _, e: self.print_error(
                f"Could not pause watcher: {e} caused by {e.__cause__}",
            ),
        )

    def do_continue(self, line: str):
        """
        do_continue Continue the watcher process after it has been paused.

        Args:
            line (str): Empty string, no arguments expected
        """
        if len(line) > 0:
            print("Invalid input. Expected no arguments.", flush=True)
            return

        self.dispatch_on_watcher(
            do_is_none=lambda _: print(
                "Cannot continue watcher, no watcher present", flush=True
            ),
            do_is_running=lambda _: print(
                "Cannot continue watcher, is not sleeping", flush=True
            ),
            do_is_sleeping=lambda self: self.watcher.go_on(),
            do_else=lambda _: print(
                "Cannot continue watcher, is not running", flush=True
            ),
            do_failure=lambda _, e: self.print_error(
                f"Could not continue watcher: {e} caused by {e.__cause__}"
            ),
        )

    def do_restart(self, line: str):
        """
        do_restart Restart the watcher process.

        Args:
            line (str): Empty string, no arguments expected
        """
        if len(line) > 0:
            print("Invalid input. Expected no arguments.", flush=True)
            return

        self.dispatch_on_watcher(
            do_is_none=lambda _: print(
                "Cannot restart watcher, no watcher present", flush=True
            ),
            do_is_sleeping=lambda _: print(
                "Cannot restart watcher, is sleeping and must be continued first",
                flush=True,
            ),
            do_is_running=lambda self: self.watcher.restart(),
            do_else=lambda _: print(
                "Cannot restart watcher, is not running", flush=True
            ),
            do_failure=lambda _, e: self.print_error(
                f"Could not restart watcher: {e} caused by {e.__cause__}"
            ),
        )

    def do_status(self, line: str):
        """
        do_status Get the current status of the watcher process.

        Args:
            line (str): Empty string, no arguments expected
        """
        if len(line) > 0:
            print("Invalid input. Expected no arguments.", flush=True)
        elif self.watcher is None:
            print("No watcher present, cannot check status", flush=True)
        else:
            print("model name", self.watcher.model_name, flush=True)
            print("current input directory", self.watcher.input_directory, flush=True)
            print("current output directory", self.watcher.output_directory, flush=True)
            print("is running:", self.watcher.is_running, flush=True)
            print("is sleeping:", self.watcher.is_sleeping, flush=True)
            print("may do work:", self.watcher.may_do_work.is_set(), flush=True)
            print(
                "is done analying: ",
                self.watcher.is_done_analyzing.is_set(),
                flush=True,
            )

    def do_get_setup_info(self, line: str):
        """
        do_get_setup_info Get information about the current setup of faunanet. If no setup information is found, the cache and config directories are listed.
        Args:
            line (str): Empty string, no arguments expected
        """
        if len(line) > 0:
            print("Invalid input. Expected no arguments.", flush=True)
        elif Path(user_config_dir(), "faunanet", "install.yml").is_file() is False:
            print(
                "cache directories: ",
                [
                    f.expanduser()
                    for f in Path(user_cache_dir()).iterdir()
                    if "faunanet" in str(f)
                ],
                flush=True,
            )
            print(
                "config directories: ",
                [
                    f.expanduser()
                    for f in Path(user_config_dir()).iterdir()
                    if "faunanet" in str(f)
                ],
                flush=True,
            )
            print("No further setup information found", flush=True)
        else:
            print(
                "cache directories: ",
                [
                    f.expanduser()
                    for f in Path(user_cache_dir()).iterdir()
                    if "faunanet" in str(f)
                ],
                flush=True,
            )
            print(
                "config directories: ",
                [
                    f.expanduser()
                    for f in Path(user_config_dir()).iterdir()
                    if "faunanet" in str(f)
                ],
                flush=True,
            )
            print(
                "Current setup: ",
                read_yaml(str(Path(user_config_dir(), "faunanet", "install.yml"))),
                flush=True,
            )

    def do_change_analyzer(self, line: str):
        """
        do_change_analyzer Change the analyzer of the watcher process. This will assert data consistency in the output folder of the old analyzer.

        Args:
            line (str): Path to a custom configuration file. The file must be a yaml file with the same structure as the default configuration file.
        """

        # add a closure to encapsulate analyzer change logic
        def run_analyzer_change(cfg: str = ""):
            config = None
            try:
                config = self.update_parameters(cfg)
            except Exception as e:
                self.print_error(
                    f"An error occured while trying to update the parameters: {e} caused by {e.__cause__}"
                )

            self.watcher.change_analyzer(
                model_name=config["Analysis"]["modelname"],
                preprocessor_config=config["Data"]["Preprocessor"],
                model_config=config["Analysis"]["Model"],
                recording_config=config["Analysis"]["Recording"],
                species_predictor_config=config["Analysis"].get(
                    "SpeciesPredictor", None
                ),
                pattern=config["Analysis"]["pattern"],
                check_time=config["Analysis"]["check_time"],
                delete_recordings=config["Analysis"]["delete_recordings"],
            )

            self.wait_for_watcher_event(
                lambda s: s.watcher.is_running, limit=20, waiting_time=3
            )

        # apply the above closure to the watcher
        self.dispatch_on_watcher(
            do_is_none=lambda _: print("No watcher present, cannot change analyzer\n"),
            do_is_sleeping=lambda _: print(
                "Cannot change analyzer, watcher is sleeping\n"
            ),
            do_is_running=lambda s: s.process_arguments(
                line,
                ["--cfg"],
                do_no_inputs=lambda s: print(
                    "Cannot change analyzer, no config file provided\n"
                ),
                do_with_inputs=lambda _, kwargs: run_analyzer_change(**kwargs),
                do_with_failure=lambda s, _, e: self.print_error(
                    f"An error occured when changing analyzer: {e}, caused by {e.__cause__}\n"
                ),
            ),
            do_else=lambda _: print("Cannot change analyzer, watcher is not running\n"),
            do_failure=lambda _, e: self.print_error(
                f"An error occured when changing analyzer: {e}, caused by {e.__cause__}\n"
            ),
        )

    def emptyline(self):
        print("\n")

    def postloop(self):
        self.handle_exit()

    def postcmd(self, stop, line):
        print("\n")
        return stop

    def cmdloop(self, intro=None):
        """
        cmdloop Override of the cmd.cmdloop method to catch KeyboardInterrupts and other exceptions that might occur during the execution of the shell.

        Args:
            intro (str, optional): Intro text when starting. Defaults to None.
        """
        while self.running:
            try:
                super().cmdloop()

                while (
                    self.watcher is not None
                    and not self.watcher.exception_queue.empty()
                ):
                    e, traceback_str = self.watcher.exception_queue.get()
                    print("An error occurred in the watcher subprocess: ", e)
                    print("Traceback: ", traceback_str)
            except KeyboardInterrupt:
                print("Execution Interrupted\n")
                print("Exiting shell...\n")
                self.running = False
                break
            except Exception as e:
                print("An error occured: ", e, " caused by: ", e.__cause__)
                print("Traceback: ")
                traceback.print_exc()
                print(
                    "If you tried to modify the watcher, make sure to restart it or exit and start a new session"
                )
            finally:
                print(self.prompt, flush=True)


def run():
    import multiprocessing

    multiprocessing.set_start_method("spawn", True)
    FaunanetCmd().cmdloop()



import shutil
import pooch
import yaml
import os
from pathlib import Path
from platformdirs import user_config_dir, user_cache_dir
from faunanet import utils
from importlib.resources import files
import faunanet

FAUNANET_HOME = None
FAUNANET_MODELS = None
FAUNANET_OUTPUT = None
FAUNANET_EXAMPLES = None
FAUNANET_CONFIG = None
FAUNANET_CACHE = None


def make_directories(base_cfg_dirs: dict):
    """
    make_directories Make all the directories for faunanet.


    Args:
        base_cfg_dirs (dict): Dictionary containing paths for the main install ("home"),
        the directory where models are stored ("models"),
        and the "output" directory to store inference results and potentially other data in ("output")
    Raises:
        KeyError: A folder given in the config does not exist

    Returns:
        tuple: created folders: (faunanet-homefolder, modelsfolder, datafolder, outputfolder, examplefolder)
    """
    if "home" not in base_cfg_dirs:
        raise KeyError("The home folder for faunanet must be given in the base config")

    if "models" not in base_cfg_dirs:
        raise KeyError(
            "The models folder for faunanet must be given in the base config"
        )

    if "output" not in base_cfg_dirs:
        raise KeyError(
            "The output folder for faunanet must be given in the base config"
        )

    ish = Path(base_cfg_dirs["home"]).expanduser().resolve()
    ism = Path(base_cfg_dirs["models"]).expanduser().resolve()
    iso = Path(base_cfg_dirs["output"]).expanduser().resolve()
    ise = (Path(base_cfg_dirs["home"]).expanduser() / Path("example")).resolve()
    iscfg = Path(user_config_dir()) / "faunanet"
    iscache = Path(user_cache_dir()) / "faunanet"

    if os.getenv("FAUNANET_TEST_MODE") == "True":
        iscfg = Path(user_config_dir()) / "faunanet_tests"
        iscache = Path(user_cache_dir()) / "faunanet_tests"

    for p in [ish, ism, iso, ise]:
        p.mkdir(parents=True, exist_ok=False)

    for p in [iscfg, iscache]:
        p.mkdir(parents=True, exist_ok=True)

    return ish, ism, iso, ise, iscfg, iscache


def download_model_files(model_dir: str = "models"):
    """
    download_model_files Download default model files for faunanet. This is required to run before anything else is done with faunanet.

    Args:
        model_dir (str, optional): Path to put the model files at. Defaults to "models".

    Raises:
        FileNotFoundError: When the model directory does not exist.
    """
    print("... Downloading model files...")

    ism = Path(model_dir)

    if ism.exists() is False:
        raise FileNotFoundError(f"The folder {model_dir} does not exist")

    # have filenames and checksum hardcoded for now. "None" indicates unknown or changing filenames
    model_file_names = {
        "birdnet_default_v2.4/model.tflite": "sha256:55f3e4055b1a13bfa9a2452731d0d34f6a02d6b775a334362665892794165e4c",
        "birdnet_custom_v2.4/model.tflite": "sha256:5eee37652430c54490321e0d2096d55817b4ac4da2301b594f03fcb9d52af741",
        "birdnet_default_v2.4/species_presence_model.tflite": "sha256:1226f23fc20362617deb09178f366111b70cf085c2c82893f2814e5acedce6c2",
        "google_bird_classification/saved_model.pb": "sha256:c6bdf0c2659f6d4713c670b92cceb083f8c1b401aee79ef6912614a1d89b6f97",
        "google_bird_classification/variables/variables.index": None,
        "google_bird_classification/variables/variables.data-00000-of-00001": "sha256:d967ea7bfaccdcfe4ff59f8d9f3bdec69181e4e79fd7abebf692100efcdfcc56",
    }

    label_file_names = {
        "birdnet_default_v2.4/labels.txt": None,
        "birdnet_custom_v2.4/labels.txt": None,
        "google_bird_classification/assets/family.csv": None,
        "google_bird_classification/assets/genus.csv": None,
        "google_bird_classification/assets/label.csv": None,
        "google_bird_classification/assets/order.csv": None,
        "google_bird_classification/labels.txt": None,
        "google_bird_classification/train_metadata.csv": None,
    }

    model_code = {
        "birdnet_default_v2.4/model.py": None,
        "birdnet_custom_v2.4/model.py": None,
        "google_bird_classification/model.py": None,
    }

    preprocessor_code = {
        "birdnet_default_v2.4/preprocessor.py": None,
        "birdnet_custom_v2.4/preprocessor.py": None,
        "google_bird_classification/preprocessor.py": None,
    }

    config_files = {
        "birdnet_default_v2.4/default.yml": None,
        "birdnet_custom_v2.4/default.yml": None,
        "google_bird_classification/default.yml": None,
    }

    models_data = pooch.create(
        path=pooch.os_cache("faunanet_downloads"),
        base_url="https://huggingface.co/MaHaWo/faunanet_test_models/resolve/main",
        registry=(
            model_file_names
            | label_file_names
            | model_code
            | preprocessor_code
            | config_files
        ),
    )

    for name in [
        "model.tflite",
        "labels.txt",
        "species_presence_model.tflite",
        "model.py",
        "preprocessor.py",
        "default.yml",
    ]:
        (ism / Path("birdnet_default")).mkdir(parents=True, exist_ok=True)
        shutil.copy(
            Path(models_data.fetch(f"birdnet_default_v2.4/{name}", progressbar=False)),
            ism / Path("birdnet_default"),
        )

    for name in [
        "model.tflite",
        "labels.txt",
        "model.py",
        "preprocessor.py",
        "default.yml",
    ]:
        (ism / Path("birdnet_custom")).mkdir(parents=True, exist_ok=True)

        shutil.copy(
            Path(models_data.fetch(f"birdnet_custom_v2.4/{name}", progressbar=False)),
            ism / Path("birdnet_custom"),
        )

    for name in [
        "saved_model.pb",
        "labels.txt",
        "train_metadata.csv",
        "model.py",
        "preprocessor.py",
        "default.yml",
    ]:
        (ism / Path("google_perch")).mkdir(parents=True, exist_ok=True)

        shutil.copy(
            Path(
                models_data.fetch(
                    f"google_bird_classification/{name}", progressbar=False
                )
            ),
            ism / Path("google_perch"),
        )

    # assets and variables need to live in separate directories to load the model correctly
    for name in [
        "variables/variables.index",
        "variables/variables.data-00000-of-00001",
    ]:

        (ism / Path("google_perch") / "variables").mkdir(parents=True, exist_ok=True)

        shutil.copy(
            Path(
                models_data.fetch(
                    f"google_bird_classification/{name}", progressbar=False
                )
            ),
            ism / Path("google_perch") / "variables",
        )

    for name in [
        "assets/family.csv",
        "assets/genus.csv",
        "assets/label.csv",
        "assets/order.csv",
    ]:

        (ism / Path("google_perch") / "assets").mkdir(parents=True, exist_ok=True)

        shutil.copy(
            Path(
                models_data.fetch(
                    f"google_bird_classification/{name}", progressbar=False
                )
            ),
            ism / Path("google_perch") / "assets",
        )


def download_example_data(example_dir: str = "examples"):
    """
    download_example_data Download the example audio files used by faunanet for its tests and examples.

    Args:
        example_dir (str, optional): Path to the faunanet example directory. Defaults to 'examples'.
    """

    print("... Downloading example files...")

    ise = Path(example_dir)

    if ise.exists() is False:
        raise FileNotFoundError(f"The folder {example_dir} does not exist")

    # create registries to pull files from. hardcode names and hashes for now
    example_data_file_names = {
        "corrupted.wav": "sha256:68cbc7c63bed90c2ad4fb7d3b5cc608c82cebeaf5e91d5e8d6793d8645b30b95",
        "soundscape.wav": "sha256:df312b45bc82ce4c638c3e9e09d748702ea14a91ec29e4e8e0676d3e3e015fd7",
        "trimmed.wav": "sha256:a4b3c078d2c617264dc5e6a0d62deedba9232699580c9c73d237496726c194b3",
        "species_list.txt": None,
    }

    example_data = pooch.create(
        path=pooch.os_cache("faunanet_downloads"),
        base_url="https://huggingface.co/datasets/MaHaWo/iSparrow_test_data/resolve/main",
        registry=example_data_file_names,
    )

    # don't leave them in the cached folder but copy to where user can actually see them.
    for name in example_data_file_names.keys():
        shutil.copy(
            Path(example_data.fetch(name, progressbar=False)),
            ise,
        )


def set_up(custom_config: str = None):
    """
    set_up Set up the faunanet directories and download the necessary data. This is required to run before anything else is done with faunanet.

    Args:
        custom_config (str, optional): Path to a custom installation file. See the 'install.yml' file provided with this package for possible customization options. Defaults to None.

    Raises:
        FileExistsError: When there is an existing faunanet installation.
    """
    print("Creating faunanet folders and downloading data... ")

    packagebase = files(faunanet)
    install_cfg = utils.read_yaml(packagebase / "install.yml")

    if custom_config is not None:
        custom_install_config = utils.read_yaml(custom_config)
        utils.update_dict_leafs_recursive(install_cfg, custom_install_config)

    for key in ["home", "models", "output"]:
        if Path(install_cfg["Directories"][key]).expanduser().resolve().exists():
            raise FileExistsError(
                f"{key} directory already exists. Please remove it before running the installation."
            )

    home, models, output, examples, config, cache = make_directories(
        install_cfg["Directories"]
    )

    if Path(config, "install.yml").exists():

        with open(Path(config) / "install.yml", "r") as yfile:
            old_cfg = yaml.safe_load(yfile)["Directories"]
            old_dirs = [Path(p).expanduser().resolve() for p in old_cfg.values()]

        for dir in [home, models, output, examples, config, cache]:

            # avoid compromising existing installations
            if Path(dir).exists() and Path(dir).expanduser().resolve() not in old_dirs:
                shutil.rmtree(dir)
        p = Path(config, "install.yml")
        raise FileExistsError(
            f"An faunanet installation already exists at {p}. Please remove it before running the installation."
        )

    # copy over config files
    with open(Path(config) / "install.yml", "w") as yfile:
        yaml.safe_dump(install_cfg, yfile)
    shutil.copy(packagebase / "default.yml", config)

    # copy over docker files 
    Path(home, "docker").mkdir(parents=True, exist_ok=True)
    shutil.copy(packagebase / Path("faunanet.dockerfile"), home / "docker")
    shutil.copy(packagebase / Path("docker-compose.yml"), home / "docker")
    shutil.copy(packagebase / Path("startup_docker.py"), home / "docker")

    download_model_files(model_dir=models.resolve())

    download_example_data(example_dir=examples.resolve())

    global FAUNANET_HOME, FAUNANET_MODELS, FAUNANET_OUTPUT, FAUNANET_EXAMPLES, FAUNANET_CACHE, FAUNANET_CONFIG

    FAUNANET_HOME = home
    FAUNANET_MODELS = models
    FAUNANET_OUTPUT = output
    FAUNANET_EXAMPLES = examples
    FAUNANET_CACHE = cache
    FAUNANET_CONFIG = config

    print("Installation finished")
